@@ -20,32 +20,34 @@
  
 #include <libgnomeui/libgnomeui.h>
 
-#include "application.h"
-#include "dvb_dvr.h"
-#include "dvb_si.h"
-#include "integer.h"
-#include "exception_handler.h"
-#include "network.h"
-#include "mutex_lock.h"
-#include "gdk_lock.h"
 #include "config.h"
-#include "ts2pes.h"
-#include "scan_dialog.h"
+#include "application.hh"
+#include "dvb_dvr.hh"
+#include "dvb_si.hh"
+#include "integer.hh"
+#include "exception_handler.hh"
+#include "network.hh"
+#include "mutex_lock.hh"
+#include "gdk_lock.hh"
+#include "scan_dialog.hh"
+#include "xine_engine.hh"
+#include "mplayer_engine.hh"
 
-#define AUTO_SURF_INTERVAL			5
-#define POKE_INTERVAL				60
 #define RECORDING_CHECK_INTERVAL	30
 #define EPG_CHECK_INTERVAL			5
 #define EPG_UPDATE_INTERVAL			60
-#define MAX_EPG_EVENT_LIST_SIZE		500
-#define BUFFER_SIZE					PACKET_SIZE*100 // Don't exceed 64K for UDP
+#define MAX_EPG_EVENT_LIST_SIZE		100
+#define MAX_NUMBER_STORED_EVENTS	10000
 
 Application* current_application = NULL;
-char* configuration_file_argument = NULL;
+gchar* configuration_file_argument = NULL;
+gchar **remaining_args = NULL;
 
 GOptionEntry option_entries[] = {
 	{ "configuration-file", 'c', G_OPTION_FLAG_FILENAME, G_OPTION_ARG_FILENAME,
 		&configuration_file_argument, "Configuration file to use", "<configuration file>" },
+	{ G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY,
+		&remaining_args, "Video files to play" },
 	{ NULL }
 };
 
@@ -57,20 +59,19 @@ Application::Application (int argc, char** argv)
 	log_file					= NULL;
 	application_quit			= false;
 	terminate_threads			= true;
-	stream_thread				= NULL;
 	epg_thread					= NULL;
 	tuner						= NULL;
 	main_window					= NULL;
 	recording_manager			= NULL;
-	recording_file_stream		= NULL;
-	network_stream				= NULL;
 	error_messages				= NULL;
 	auto_surf_iterator			= NULL;
 	epg_events					= NULL;
 	epg_event_list_size			= 0;
 	scheduled_recording_node	= NULL;
 	channel_change_complete		= false;
-
+	engine						= NULL;
+	streamer					= NULL;
+	
 	g_static_rec_mutex_init (&stream_mutex);
 	g_static_rec_mutex_init (&configuration_mutex);
 	g_static_rec_mutex_init (&error_message_mutex);
@@ -109,6 +110,18 @@ Application::Application (int argc, char** argv)
 Application::~Application()
 {
 	stop();
+	
+	if (streamer != NULL)
+	{
+		delete streamer;
+		streamer = NULL;
+	}
+
+	if (engine != NULL)
+	{
+		delete engine;
+		engine = NULL;
+	}
 
 	if (main_window != NULL)
 	{
@@ -127,6 +140,12 @@ Application::~Application()
 		tuner = NULL;
 		delete tuner;
 	}
+	
+	if (engine != NULL)
+	{
+		delete engine;
+		engine = NULL;
+	}
 }
 
 Application& Application::get_current()
@@ -172,7 +191,6 @@ void Application::init()
 	g_free(exe_dir);
 	
 	log_file = new IO::Channel::Channel(application_directory + "/me-tv.log", O_CREAT | O_TRUNC | O_WRONLY);
-	log_file->set_encoding(NULL);
 	
 	Log::write("Me TV Version: %s", PACKAGE_VERSION);
 
@@ -187,36 +205,50 @@ void Application::init()
 	}
 	configuration.load(configuration_file);
 	
+	if (remaining_args != NULL)
+	{
+		g_strv_length (remaining_args);
+		video_file = remaining_args[0];
+		g_strfreev (remaining_args);
+		remaining_args = NULL;
+	}
+
 	String recording_directory(g_get_home_dir());
 	
 	configuration.set_default_string_value("dvr_path", "/dev/dvb/adapter0/dvr0");
 	configuration.set_default_string_value("demux_path", "/dev/dvb/adapter0/demux0");
 	configuration.set_default_string_value("frontend_path", "/dev/dvb/adapter0/frontend0");
+	configuration.set_default_string_value("ca_path", "/dev/dvb/adapter0/ca0");
 	configuration.set_default_boolean_value("mute", false);
 	configuration.set_default_string_value("engine_type", "xine");
 	configuration.set_default_string_value("recording_directory", recording_directory);
-	configuration.set_default_string_value("recording_type", "ts");
 	configuration.set_default_int_value("video_position", -1);
 	configuration.set_default_string_value("epg_path", get_directory() + "/epg.xml");
 	configuration.set_default_string_value("channel_file_path", get_directory() + "/channels.conf");
-	configuration.set_default_int_value("x", -1);
-	configuration.set_default_int_value("y", -1);
-	configuration.set_default_int_value("width", -1);
-	configuration.set_default_int_value("height", -1);
+	configuration.set_default_int_value("x", 100);
+	configuration.set_default_int_value("y", 100);
+	configuration.set_default_int_value("width", 500);
+	configuration.set_default_int_value("height", 400);
 	configuration.set_default_int_value("record_extra_before", 5);
 	configuration.set_default_int_value("record_extra_after", 10);
+	configuration.set_default_string_value("broadcast_address", "192.168.0.255");
 	configuration.set_default_int_value("broadcast_port", 2005);
 	configuration.set_default_boolean_value("always_on_top", true);
 	configuration.set_default_boolean_value("fullscreen", false);
 	configuration.set_default_boolean_value("show_controls", true);
-	configuration.set_default_boolean_value("show_epg", true);
+	configuration.set_default_boolean_value("show_epg", false);
+	configuration.set_default_boolean_value("show_channel_selector", false);
 	configuration.set_default_boolean_value("quit_on_close", true);
 	configuration.set_default_boolean_value("start_minimised_in_tray", false);	
 	configuration.set_default_string_value("epg_encoding", "auto");
 	configuration.set_default_boolean_value("fullscreen_workaround", false);
-	configuration.set_default_boolean_value("pid_selection_workaround", false);
 	configuration.set_default_int_value("epg_span_hours", 3);
-	configuration.set_default_boolean_value("sort_channels_alphabetically", false);
+	configuration.set_default_int_value("channel_sort", 0);
+	configuration.set_default_string_value("lnb_type", "none");
+	configuration.set_default_boolean_value("pid_workaround", false);
+	configuration.set_default_string_value("preferred_language", "");
+	configuration.set_default_int_value("auto_surf_interval", 10);
+
 	configuration.set_default_string_value("xine.video_driver", "auto");
 	configuration.set_default_string_value("xine.audio_driver", "auto");
 	configuration.set_default_string_value("xine.deinterlace_type", "default");
@@ -263,10 +295,9 @@ void Application::init()
 	}
 	else
 	{
-		String test_path = get_test_path();
-		if (!IO::File::exists(test_path))
+		if (video_file.get_character_length() == 0)
 		{
-			throw Exception(_("No tuner device"));
+			throw Exception(_("No tuner device"));		
 		}
 	}
 	
@@ -283,13 +314,9 @@ void Application::init()
 	{
 		throw Exception(_("There are no usable channels in the channels.conf file."));
 	}
-
+	
 	recording_manager = new RecordingManager();
 	
-	main_window = new MainWindow();
-
-	mute(configuration.get_boolean_value("mute"));
-
 	if (configuration.has_value("last_channel"))
 	{
 		String last_channel_name = configuration.get_string_value("last_channel");
@@ -306,6 +333,29 @@ void Application::init()
 		Channel* channel = (Channel*)channels->data;
 		video_channel_name = channel->name;		
 	}
+
+	main_window = new MainWindow();
+
+	String engine_type = configuration.get_string_value("engine_type");
+	Log::write(_("Using '%s' engine"), engine_type.c_str());
+	if (engine_type == "xine")
+	{
+		engine = new XineEngine();
+	}
+	else if (engine_type == "mplayer")
+	{
+		engine = new MPlayerEngine();
+	}
+	else if (engine_type == "none")
+	{
+		// No engine
+	}
+	else
+	{
+		throw Exception(_("Unknown engine type '%s'"), engine_type.c_str());
+	}
+
+	mute(configuration.get_boolean_value("mute"));
 	
 	dialog_about = glade.get_widget("dialog_about");
 	g_signal_connect( G_OBJECT ( dialog_about ), "delete-event", G_CALLBACK ( gtk_widget_hide ), this );
@@ -315,13 +365,18 @@ void Application::init()
 	
 	if (!configuration.get_boolean_value("start_minimised_in_tray"))
 	{
-		main_window->show();
+		if (main_window != NULL)
+		{
+			main_window->show();
+		}
 	}
 	else
 	{
 		mute(true);
 	}
 
+	streamer = new Streamer();
+	
 	timeout_id = gdk_threads_add_timeout (1000, on_timer, this);
 	
 	while (!application_quit)
@@ -346,6 +401,11 @@ RecordingManager& Application::get_recording_manager()
 	return *recording_manager;
 }
 
+String Application::get_video_file() const
+{
+	return video_file;
+}
+
 void Application::quit()
 {
 	gboolean want_close = true;
@@ -363,7 +423,6 @@ void Application::quit()
 	{
 		application_quit = true;
 		g_source_remove(timeout_id);
-		stop();
 		gtk_main_quit();
 	}
 }
@@ -383,8 +442,10 @@ void Application::check_error_messages()
 	}
 }
 
-void Application::push_epg_event(const EpgEvent& event)
+gboolean Application::push_epg_event(const EpgEvent& event)
 {
+	gboolean result = true;
+	
 	MutexLock lock(&epg_mutex);
 
 	static gboolean message_recorded = false;
@@ -396,21 +457,26 @@ void Application::push_epg_event(const EpgEvent& event)
 	}
 	else
 	{		
+		result = false;
+		
 		if (!message_recorded)
 		{
 			message_recorded = true;
 			Log::write(_("Too many events in EPG queue"));
 		}
 	}
+	
+	return result;
 }
 
-void Application::check_epg_events()
+gboolean Application::check_epg_events()
 {
-	MutexLock lock(&epg_mutex);
 	gboolean update = false;
 	
 	while (epg_events != NULL)
-	{		
+	{
+		MutexLock lock(&epg_mutex);
+	
 		EpgEvent* event = (EpgEvent*)epg_events->data;
 			
 		if (epg.add_event(*event) > 0)
@@ -419,7 +485,7 @@ void Application::check_epg_events()
 		}
 
 		delete event;
-		epg_events = g_slist_remove_link(epg_events, epg_events);
+		epg_events = g_slist_delete_link(epg_events, epg_events);
 		epg_event_list_size--;
 	}
 	
@@ -428,16 +494,7 @@ void Application::check_epg_events()
 		update = true;
 	}
 	
-	epg_event_list_size = 0;
-
-	if (update == true)
-	{
-		Log::write(_("Saving EPG data"));
-		epg.save();
-				
-		Log::write(_("Updating EPG UI"));
-		update_epg();
-	}
+	return update;
 }
 
 gboolean Application::on_check_error_messages_timer(gpointer data)
@@ -452,11 +509,11 @@ gboolean Application::on_check_error_messages_timer(gpointer data)
 
 gboolean Application::on_timer(gpointer data)
 {
-	static time_t last_poke_time = 0;
 	static time_t last_auto_surf_time = 0;
 	static time_t last_epg_check_time = 0;
 	static time_t last_epg_update_time = 0;
 	static gboolean initialise = true;
+	static gboolean do_auto_surf_return_channel = false;
 	
 	Application& application = *(Application*)data;
 	MutexLock lock(&application.stream_mutex);
@@ -467,28 +524,37 @@ gboolean Application::on_timer(gpointer data)
 	{
 		initialise = false;
 		application.update_epg();
-		application.change_channel(application.get_video_channel());
+		if (application.video_file.get_character_length() == 0)
+		{
+			application.change_channel(application.get_video_channel());
+		}
+		else
+		{
+			application.start();
+		}
+	}
+		
+	if (application.video_file.get_character_length() == 0 && application.channel_change_complete)
+	{
+		if (application.main_window != NULL)
+		{
+			application.main_window->channel_change_complete();
+		}
+		application.channel_change_complete = false;
 	}
 	
-	if (application.channel_change_complete)
+	if (application.main_window != NULL)
 	{
-		application.main_window->channel_change_complete();
-		application.channel_change_complete = false;
+		application.main_window->on_timer();
 	}
 	
 	time_t now = DateTime::now(); 
 
-	application.main_window->on_timer();
-	
-	if (now - last_poke_time > POKE_INTERVAL)
+	if (application.engine != NULL)
 	{
-		if (application.main_window->is_visible() && !application.main_window->is_minimised()) 		
-		{ 		
-			application.main_window->send_fake_key();
-		}
-		last_poke_time = now;
+		application.engine->on_timer();
 	}
-
+	
 	if (now - application.last_recording_check_time > RECORDING_CHECK_INTERVAL)
 	{
 		xmlNodePtr node = application.get_recording_manager().check();
@@ -506,7 +572,7 @@ gboolean Application::on_timer(gpointer data)
 		}
 		else
 		{
-			if (application.is_recording() && application.scheduled_recording_node != NULL)
+			if (application.streamer->is_recording() && application.scheduled_recording_node != NULL)
 			{
 				application.scheduled_recording_node = NULL;
 				application.record(false);
@@ -516,26 +582,38 @@ gboolean Application::on_timer(gpointer data)
 		application.last_recording_check_time = now;
 	}
 	
-	if ((application.auto_surf_iterator != NULL) && (now - last_auto_surf_time > AUTO_SURF_INTERVAL))
+	guint auto_surf_interval = application.configuration.get_int_value("auto_surf_interval");
+	if ((application.auto_surf_iterator != NULL) && (now - last_auto_surf_time > auto_surf_interval))
 	{
 		Channel& channel = *(Channel*)application.auto_surf_iterator->data;
 		application.change_channel(channel, true);
 		application.auto_surf_iterator = g_slist_next(application.auto_surf_iterator);
+		do_auto_surf_return_channel = application.auto_surf_iterator == NULL;
+		last_auto_surf_time = now;
+	}
 		
-		if (application.auto_surf_iterator == NULL)
+	if ((now - last_auto_surf_time > auto_surf_interval) && do_auto_surf_return_channel)
+	{
+		if (application.main_window != NULL)
 		{
 			application.main_window->auto_surf(false);
-			application.change_channel(application.auto_surf_return_channel);
 		}
-		
-		last_auto_surf_time = now;
+		application.change_channel(application.auto_surf_return_channel);
+		do_auto_surf_return_channel = false;
 	}
 	
 	if (application.tuner != NULL)
 	{
 		if (now - last_epg_check_time > EPG_CHECK_INTERVAL)
 		{
-			application.check_epg_events();
+			if (application.check_epg_events())
+			{
+				Log::write(_("Saving EPG data"));
+				application.epg.save();
+						
+				Log::write(_("Updating EPG UI"));
+				application.update_epg();
+			}
 			last_epg_check_time = now;
 		}
 	}
@@ -551,7 +629,7 @@ gboolean Application::on_timer(gpointer data)
 	return TRUE;
 }
 
-void Application:: start_epg_thread()
+void Application::start_epg_thread()
 {
 	MutexLock lock(&stream_mutex);
 
@@ -574,19 +652,8 @@ void Application::start()
 	MutexLock lock(&stream_mutex);
 	
 	terminate_threads = false;
-
-	if (main_window != NULL && main_window->get_engine() != NULL)
-	{
-		main_window->get_engine()->open();
-	}
-	
-	stream_thread = g_thread_create((GThreadFunc)stream_thread_function, this, TRUE, NULL);
-	if (stream_thread == NULL)
-	{
-		throw SystemException(_("Failed to create stream thread"));
-	}
-	Log::write(_("Stream thread created"));
-	
+		
+	streamer->start();	
 	start_epg_thread();
 }
 
@@ -595,7 +662,6 @@ void Application::stop()
 	terminate_threads = true;
 	
 	MutexLock lock(&stream_mutex);
-	record(false);
 	
 	if (epg_thread != NULL)
 	{
@@ -603,24 +669,18 @@ void Application::stop()
 		epg_thread = NULL;
 	}
 	
+	if (streamer != NULL)
+	{
+		record(false);
+		streamer->stop();
+	}
+
 	if (tuner != NULL && !video_channel_name.is_empty())
 	{
 		Log::write(_("Purging remaining EPG events"));
 		check_epg_events();
 		Log::write(_("EPG events purged"));
 	}
-
-	if (stream_thread != NULL)
-	{
-		SCOPE_LOG(_("Stopping stream thread"));
-		g_thread_join(stream_thread);
-		stream_thread = NULL;
-	}
-	
-	if (main_window != NULL && main_window->get_engine() != NULL)
-	{
-		main_window->get_engine()->close();
-	}
 }
 
 String Application::get_video_event_title()
@@ -683,13 +743,15 @@ void Application::change_channel (Channel& channel, gboolean keep_auto_surf)
 {
 	Log::write(_("Request to change channel to '%s'"), channel.name.c_str());
 
-	if (keep_auto_surf == false)
+	if (auto_surf_iterator != NULL && keep_auto_surf == false)
 	{
-		auto_surf_iterator = NULL;
+		auto_surf(false);
 	}
 	
 	MutexLock lock(&stream_mutex);
 	
+	video_file = "";
+	
 	if (video_channel_name == channel.name && terminate_threads == false)
 	{
 		Log::write(_("Already tuned to '%s'"), channel.name.c_str());
@@ -699,7 +761,7 @@ void Application::change_channel (Channel& channel, gboolean keep_auto_surf)
 		bool want_change = true;
 		bool same_transponder = false;
 		
-		if (recording_file_stream != NULL && !same_transponder)
+		if (streamer->is_recording() && !same_transponder)
 		{
 			gint response = show_message_dialog(
 				_("You have to stop recording to change to this channel. Do you still want to change channels?"),
@@ -717,8 +779,17 @@ void Application::change_channel (Channel& channel, gboolean keep_auto_surf)
 		{
 			Log::write(_("Changing channel to '%s'"), channel.name.c_str());
 			
+			if (main_window != NULL)
+			{
+				main_window->set_dual_language_state(ENGINE_DUAL_LANGUAGE_STATE_DISABLED);
+			}
+			else if (engine != NULL)
+			{
+				engine->set_dual_language_state(ENGINE_DUAL_LANGUAGE_STATE_DISABLED);
+			}
+
 			stop();
-			
+				
 			if (tuner != NULL)
 			{
 				Transponder& transponder = channel.get_transponder();
@@ -731,7 +802,10 @@ void Application::change_channel (Channel& channel, gboolean keep_auto_surf)
 
 			start();
 
-			main_window->change_channel_start();
+			if (main_window != NULL)
+			{
+				main_window->change_channel_start();
+			}
 			
 			Log::write(_("Channel changed to '%s'"), channel.name.c_str());
 		}
@@ -745,30 +819,30 @@ void Application::toggle_mute()
 
 void Application::broadcast(gboolean state)
 {
-	if (!state && network_stream == NULL)
+	if (!state && !streamer->is_broadcasting())
 	{
 		return;
 	}
 
-	if (state && network_stream != NULL)
+	if (state && streamer->is_broadcasting())
 	{
 		return;
 	}
 
 	MutexLock lock(&stream_mutex);
-	Log::write(state ? _("Broadcast on") : _("Broadcast off"));
-	main_window->broadcast(state);
 
 	if (state)
 	{
-		network_stream = new MulticastSocket("224.0.1.2",
-			configuration.get_int_value("broadcast_port"));
+		streamer->enable_broadcasting();
 	}
 	else
 	{
-		MulticastSocket* filestream = network_stream;
-		network_stream = NULL;
-		delete filestream;
+		streamer->disable_broadcasting();
+	}
+
+	if (main_window != NULL)
+	{
+		main_window->broadcast(state);
 	}
 }
 
@@ -780,6 +854,7 @@ void Application::toggle_auto_surf()
 void Application::auto_surf(gboolean state)
 {
 	MutexLock lock(&stream_mutex);
+
 	gboolean in_auto_surf = auto_surf_iterator != NULL;
 	
 	if (state && in_auto_surf)
@@ -796,7 +871,7 @@ void Application::auto_surf(gboolean state)
 
 	if (state && !in_auto_surf)
 	{
-		if (recording_file_stream != NULL)
+		if (streamer->is_recording())
 		{
 			throw Exception(_("Can't auto surf while recording"));
 		}
@@ -804,47 +879,54 @@ void Application::auto_surf(gboolean state)
 		{
 			auto_surf_return_channel = get_video_channel().name;
 			auto_surf_iterator = get_channel_manager().get_channels();
-			main_window->auto_surf(true);
+			if (main_window != NULL)
+			{
+				main_window->auto_surf(true);
+			}
 		}
 	}
 	
 	if (!state && in_auto_surf)
 	{
 		auto_surf_iterator = NULL;
-		main_window->auto_surf(false);
+		if (main_window != NULL)
+		{
+			main_window->auto_surf(false);
+		}
 	}
 }
 
 void Application::mute(gboolean state)
 {
 	MutexLock lock(&stream_mutex);
-	main_window->mute(state);
+	if (main_window != NULL)
+	{
+		main_window->mute(state);
+	}
+	
+	if (engine != NULL)
+	{
+		engine->mute(state);
+	}
 	configuration.set_boolean_value("mute", state);
 }
 
 void Application::toggle_broadcast()
 {
-	broadcast(network_stream == NULL);
+	broadcast(streamer->is_broadcasting());
 }
 
 void Application::hide_controls()
 {
-	main_window->hide_controls();
-}
-
-void Application::audio_channel(int channel)
-{
-	main_window->audio_channel(channel);
-}
-
-void Application::subtitle_channel(int channel)
-{
-	main_window->subtitle_channel(channel);
+	if (main_window != NULL)
+	{
+		main_window->hide_controls();
+	}
 }
 
 void Application::toggle_record()
 {
-	record(!is_recording());
+	record(!streamer->is_recording());
 }
 
 void Application::start_record(const String& title)
@@ -854,8 +936,25 @@ void Application::start_record(const String& title)
 
 String Application::fix_path(const String& path)
 {
-	String result = path;
-	result.canon(G_CSET_A_2_Z G_CSET_a_2_z G_CSET_DIGITS "-_", ' ');	
+	String result;
+	const char* next = path.c_str();
+	if (next != NULL)
+	{
+		gunichar ch = g_utf8_get_char(next);
+		while (ch != 0)
+		{
+			if (g_unichar_isalnum(ch) || ch == '_' || ch == '-' || ch == '\''  || ch == ',')
+			{
+				result += ch;
+			} else {
+				result += ' ';
+			}
+			
+			next = g_utf8_find_next_char(next, NULL);
+			ch = g_utf8_get_char(next);
+		}
+	}
+	
 	return result;
 }
 
@@ -865,14 +964,15 @@ void Application::stop_record()
 	
 	gboolean was_scheduled = scheduled_recording_node != NULL;
 	
-	main_window->record(false);
-
-	if (recording_file_stream != NULL)
+	if (main_window != NULL)
+	{
+		main_window->record(false);
+	}
+	
+	if (streamer->is_recording())
 	{
 		scheduled_recording_node = NULL;
-		IO::Channel* filestream = recording_file_stream;
-		recording_file_stream = NULL;
-		delete filestream;
+		streamer->stop_recording();
 	}
 	
 	if (was_scheduled && !application_quit) 		
@@ -891,15 +991,18 @@ void Application::stop_record()
 
 void Application::start_record_force(Channel& channel, const String& title)
 {
-	if (get_video_channel().name != channel.name)
+	if (get_video_channel().name != channel.name || !streamer->is_recording())
 	{
-		if (recording_file_stream != NULL)
+		if (get_video_channel().name != channel.name)
 		{
-			stop_record();
+			if (streamer->is_recording())
+			{
+				stop_record();
+			}
+			change_channel(channel);
 		}
-		change_channel(channel);
+		start_record(channel, title);
 	}
-	start_record(channel, title);
 }
 
 void Application::start_record(Channel& channel, const String& title)
@@ -932,22 +1035,23 @@ void Application::start_record(Channel& channel, const String& title)
 	Log::write(_("Recording filename is '%s'"), filename.c_str());
 	recording_file_name = filename;
 
-	recording_file_stream = new IO::Channel(recording_file_name, O_CREAT | O_WRONLY);
-	recording_file_stream->set_encoding(NULL);
-	Log::write(_("Created recording file stream"));
+	streamer->start_recording(recording_file_name);
 	
-	main_window->record(true);
+	if (main_window != NULL)
+	{
+		main_window->record(true);
+	}
 	Log::write(_("Recording started"));
 }
 
 void Application::record(gboolean state)
 {
-	if (state && recording_file_stream != NULL)
+	if (state && streamer->is_recording())
 	{
 		return;
 	}
 
-	if (!state && recording_file_stream == NULL)
+	if (!state && !streamer->is_recording())
 	{
 		return;
 	}
@@ -961,7 +1065,7 @@ void Application::record(gboolean state)
 		Log::write(_("Request to stop recording"));
 	}
 
-	if (recording_file_stream != NULL)
+	if (streamer->is_recording())
 	{
 		stop_record();
 	}
@@ -996,289 +1100,87 @@ gpointer Application::epg_thread_function(Application* application)
 {
 	TRY;
 
-	gboolean is_atsc = application->tuner != NULL && application->tuner->get_frontend_type() == FE_ATSC;
-	DvbDemuxer demuxer_eit(application->demux_path);
+	guint stored_events[MAX_NUMBER_STORED_EVENTS];
+	gsize stored_events_count = 0;
+	
 	const Transponder& transponder = application->get_current_transponder();
-	demuxer_eit.set_buffer_size(DVB_SECTION_BUFFER_SIZE);
-	demuxer_eit.set_filter(EIT_PID, EIT_ID);
+	DvbDemuxer demuxer_eit(application->demux_path);
+	gboolean is_atsc = application->tuner->get_frontend_type() == FE_ATSC;
+	if (is_atsc)
+	{
+		demuxer_eit.set_filter(0x1FFB, EIT_ID, 0);
+	}
+	else
+	{
+		demuxer_eit.set_filter(EIT_PID, EIT_ID, 0);
+	}
 	DvbSectionParser parser;
 		
-	while (!application->terminate_threads)
+	gboolean done = application->terminate_threads;
+	while (!done)
 	{
 		try
 		{
-			DvbEventInformationSection section;
+			EventInformationSection section;
 
 			parser.parse_eis(demuxer_eit, section);
-	
-			guint service_id = section.service_id;
 			
-			// For ATSC there's no service ID
 			if (is_atsc)
 			{
-				service_id = 0;
+				Log::write("EIT read");
 			}
+	
+			guint service_id = section.service_id;
+			guint frequency = transponder.get_frequency();
 			
-			Channel* channel = application->channel_manager.get_channel(transponder.get_frequency(), service_id);
+			Channel* channel = application->channel_manager.get_channel(frequency, service_id);
 			
 			if (channel != NULL)
 			{
-				for( unsigned int k = 0; section.short_events.size() > k; k++ )
+				for( unsigned int k = 0; section.events.size() > k; k++ )
 				{
-					EpgEvent event(channel, section.short_events[k]);
-					application->push_epg_event(event);
+					EpgEvent event(channel, section.events[k]);
+
+					// Try to minimise the impact of repeated events
+					gboolean found = false;
+					guint event_id = event.get_event_id();
+					guint index = 0;
+					while (!found && index < stored_events_count)
+					{
+						if (stored_events[index] == event_id)
+						{
+							found = true;
+						}
+						
+						index++;
+					}
+					
+					if (!found)
+					{
+						if (application->push_epg_event(event) && stored_events_count < MAX_NUMBER_STORED_EVENTS)
+						{
+							stored_events[stored_events_count++] = event_id;
+						}
+					}
 				}
 			}
 		}
+		catch(const TimeoutException& ex)
+		{
+			Log::write(_("Timeout in EPG thread: %s"), ex.get_message().c_str());
+			done = true;
+		}
 		catch(const Exception& ex)
 		{
 			Log::write(_("Exception in EPG thread: %s"), ex.get_message().c_str());
 		}
+		
+		done |= application->terminate_threads;
 	}
 
 	THREAD_CATCH;
-	
-	return NULL;
-}
 
-void Application::write(uint8_t* buffer, int size)
-{
-	if (recording_file_stream != NULL)
-	{
-		recording_file_stream->write((gchar*)buffer, size);
-	}
-}
-
-void Application::on_write(uint8_t *buf,  int size, void *priv)
-{
-	Application::get_current().write(buf, size);
-}
-
-void Application::stream_loop(IO::Channel& input, guint audio_pid, guint video_pid)
-{
-	gsize length = 0;
-	gchar buffer[BUFFER_SIZE];
-	Engine* engine = main_window->get_engine();
-	gboolean record_as_ps = false;
-	
-	String recording_type = configuration.get_string_value("recording_type");
-
-	Ts2Pes ts2pes(&on_write, audio_pid, video_pid, recording_type == "ps");
-
-	if (recording_type == "ps")
-	{
-		record_as_ps = true;
-		ts2pes.go();
-	}
-	
-	while (!terminate_threads)
-	{
-		try
-		{
-			length = input.read(buffer, BUFFER_SIZE);
-			
-			if (length > 0)
-			{
-				if (engine != NULL)
-				{
-					engine->write(buffer, length);
-				}
-
-				if (recording_file_stream != NULL)
-				{
-					if (record_as_ps)
-					{
-						ts2pes.run((uint8_t*)buffer, length);
-					}
-					else
-					{
-						recording_file_stream->write(buffer, length);
-					}
-				}
-
-				if (network_stream != NULL)
-				{
-					network_stream->write(buffer, length);
-				}
-			}
-			else
-			{
-				usleep(100000);
-			}
-		}
-		catch(const Exception& ex)
-		{
-			Log::write("Exception in %s: %s", __PRETTY_FUNCTION__, ex.get_message().c_str());
-		}
-	}
-}
-
-gpointer Application::stream_thread_function(Application* application)		
-{
-	TRY;
-		
-	Channel& channel = application->get_video_channel();
-
-	SCOPE_LOG("Stream thread function");
-	if (application->tuner != NULL)
-	{
-		DvbDvr dvr(application->dvr_path);
-		DvbDemuxer demuxer_pat(application->demux_path);
-		DvbDemuxer demuxer_pmt(application->demux_path);
-		DvbDemuxer demuxer_video(application->demux_path);
-		DvbDemuxer demuxer_audio(application->demux_path);
-		DvbDemuxer demux_subtitle(application->demux_path);
-		DvbDemuxer demux_teletext(application->demux_path);
-		DvbDemuxer demux_ac3(application->demux_path);
-		DvbSectionParser parser;
-		
-		gint video_pid = channel.default_video_pid;
-		gint audio_pid = channel.default_audio_pid;
-		gint ac3_pid = 0;
-		gint subtitle_pid = channel.default_subtitle_pid;
-		guint subtitle_channel = -2;
-
-		gboolean pid_selection_workaround = application->get_configuration().get_boolean_value("pid_selection_workaround");
-		if (!pid_selection_workaround)
-		{
-			try
-			{
-				Log::write(_("Trying to auto detect PIDs"));
-				
-				ProgramAssociationSection pas;
-				demuxer_pat.set_filter(PAT_PID, PAT_ID);
-				parser.parse_pas(demuxer_pat, pas);
-				guint length = pas.program_associations.size();
-				gint program_map_pid = -1;
-				
-				Log::write(_("Found %d associations"), length);
-				for (guint i = 0; i < length; i++)
-				{
-					ProgramAssociation program_association = pas.program_associations[i];
-					
-					if (program_association.program_number == channel.service_id)
-					{
-						program_map_pid = program_association.program_map_pid;
-					}
-				}
-				
-				if (program_map_pid == -1)
-				{
-					throw Exception("Failed to find PMT ID for service");
-				}
-				
-				demuxer_pmt.set_filter(program_map_pid, PMT_ID);
-				
-				ProgramMapSection pms;
-				parser.parse_pms(demuxer_pmt, pms);
-
-				demuxer_pmt.set_pes_filter(program_map_pid, DMX_PES_OTHER);
-				demuxer_pat.set_pes_filter(PAT_ID, DMX_PES_OTHER);
-
-				if (pms.video_pids.size() > 0)
-				{
-					video_pid = pms.video_pids[0];
-				}
-				
-				gsize audio_pids_size = pms.audio_pids.size();
-				if (audio_pids_size > 0)
-				{
-					gboolean default_audio_is_available = false;
-					
-					channel.audio_pids.clear();
-					for (guint i = 0; i < audio_pids_size; i++)
-					{
-						gint pid = pms.audio_pids[i];
-						default_audio_is_available |= (pid == audio_pid);
-						channel.audio_pids.push_back(pid);
-					}
-
-					// If the default is not availablen then pick the first one that is.
-					if (!default_audio_is_available)
-					{
-						audio_pid = pms.audio_pids[0];
-					}
-				}
-				
-				gsize subtitle_pids_size = pms.subtitle_pids.size();
-				if (subtitle_pids_size > 0)
-				{
-					gboolean default_subtitle_is_available = false;
-					
-					channel.subtitle_pids.clear();
-					for (guint i = 0; i < subtitle_pids_size; i++)
-					{
-						gint pid = pms.subtitle_pids[i];
-						default_subtitle_is_available |= (pid == subtitle_pid);
-						channel.subtitle_pids.push_back(pid);
-					}
-					
-					// If the default is not available then disable subtitles.
-					if (!default_subtitle_is_available)
-					{
-						channel.default_subtitle_pid = 0;
-						subtitle_pid = 0;
-					}
-				}
-									
-				if (pms.teletext_pids.size() > 0)
-				{
-					gint teletext_pid = pms.teletext_pids[0];
-					demux_teletext.set_pes_filter(teletext_pid, DMX_PES_OTHER);
-					Log::write(_("Stream Thread: Set teletext PID filter to %d (0x%X)"), teletext_pid, teletext_pid);
-				}
-				
-				if (pms.ac3_pids.size() > 0)
-				{
-					ac3_pid = pms.ac3_pids[0];
-					demux_subtitle.set_pes_filter(pms.ac3_pids[0], DMX_PES_OTHER);
-					Log::write(_("Stream Thread: Set AC3 PID filter to %d (0x%X)"), ac3_pid, ac3_pid);
-				}
-				
-				Log::write(_("PIDs read successfully"));
-			}
-			catch(const Exception& ex)
-			{
-				Log::write(_("Failed to set PIDs automatically: %s"), ex.get_message().c_str());
-			}
-		}
-
-		demuxer_video.set_pes_filter(video_pid, DMX_PES_VIDEO);
-		Log::write(_("Stream Thread: Set video PID filter to %d (0x%X)"), video_pid, video_pid);
-	
-		if (audio_pid != 0)
-		{
-			demuxer_audio.set_pes_filter(audio_pid, DMX_PES_AUDIO);
-			Log::write(_("Stream Thread: Set audio PID filter to %d (0x%X)"), audio_pid, audio_pid);
-		}
-		else
-		{
-			audio_pid = ac3_pid;
-		}
-		
-		if (subtitle_pid != 0)
-		{
-			subtitle_channel = 0;
-			demux_subtitle.set_pes_filter(subtitle_pid, DMX_PES_OTHER);
-			Log::write(_("Stream Thread: Set subtitle PID filter to %d (0x%X)"), subtitle_pid, subtitle_pid);
-		}
-		
-		application->channel_change_complete = true;
-		dvr.set_encoding(NULL);
-		application->stream_loop(dvr, audio_pid, video_pid);
-	}
-	else
-	{
-		String test_path = application->get_test_path();
-		if (IO::File::exists(test_path))
-		{
-			IO::Channel input(test_path, O_RDONLY);
-			input.set_encoding(NULL);
-			application->stream_loop(input, channel.default_audio_pid, channel.default_video_pid);
-		}
-	}
-
-	THREAD_CATCH;
+	Log::write(_("Exiting EPG thread"));
 	
 	return NULL;
 }
@@ -1301,12 +1203,18 @@ Channel& Application::get_channel(int frequency, int service_id)
 
 void Application::toggle_fullscreen()
 {
-	main_window->toggle_fullscreen();
+	if (main_window != NULL)
+	{
+		main_window->toggle_fullscreen();
+	}
 }
 
-void Application::toggle_controls()
+void Application::toggle_epg()
 {
-	main_window->toggle_controls();
+	if (main_window != NULL)
+	{
+		main_window->toggle_epg();
+	}
 }
 
 void Application::push_error_message(const String& message)
@@ -1321,7 +1229,10 @@ void Application::push_error_message(const String& message)
 void Application::show_program_details_dialog (const String& channel_name, int program_id)
 {
 	EpgEvent event(epg.get_event(channel_name, program_id));	
-	main_window->show_program_details_dialog(event);
+	if (main_window != NULL)
+	{
+		main_window->show_program_details_dialog(event);
+	}
 }
 
 void Application::update_epg()
@@ -1330,21 +1241,18 @@ void Application::update_epg()
 	status_message = "Me TV - " + status_message;
 	status_icon->set_title(status_message);
 	
-	main_window->update_epg();
-}
-
-String Application::get_test_path() const
-{
-	gchar* test_path = g_build_filename(get_directory().c_str(), "test.ts", NULL);
-	String result = test_path;
-	g_free(test_path);
-
-	return result;
+	if (main_window != NULL)
+	{
+		main_window->update_epg();
+	}
 }
 
 void Application::show_scheduled_recording_dialog()
 {
-	main_window->show_scheduled_recording_dialog();
+	if (main_window != NULL)
+	{
+		main_window->show_scheduled_recording_dialog();
+	}
 }
 
 void Application::show_scheduled_recording_dialog(const String& description)
@@ -1356,7 +1264,10 @@ void Application::show_scheduled_recording_dialog(const String& description)
 	}
 
 	ScheduledRecording scheduled_recording(node);
-	main_window->show_scheduled_recording_dialog(scheduled_recording);
+	if (main_window != NULL)
+	{
+		main_window->show_scheduled_recording_dialog(scheduled_recording);
+	}
 }
 
 String Application::get_status_message()
@@ -1374,14 +1285,18 @@ String Application::get_status_message()
 
 void Application::toggle_visibility()
 {
-	if (main_window->is_visible())
+	if (main_window != NULL)
 	{
-		main_window->hide();
-	}
-	else
-	{
-		mute(false);
-		main_window->show();
+		if (main_window->is_visible())
+		{
+			mute(true);
+			main_window->hide();
+		}
+		else
+		{
+			mute(false);
+			main_window->show();
+		}
 	}
 }
 
@@ -1466,7 +1381,10 @@ void Application::show_scheduled_recordings_dialog()
 
 void Application::set_selected_event(xmlNodePtr event)
 {
-	main_window->set_selected_event(event);
+	if (main_window != NULL)
+	{
+		main_window->set_selected_event(event);
+	}
 }
 
 void Application::restart_stream()
@@ -1498,21 +1416,27 @@ void Application::remove_current_scheduled_recording()
 	recording_manager->remove_scheduled_recording(scheduled_recording.description);
 }
 
-DvbTuner& Application::get_tuner()
+DvbTuner* Application::get_tuner()
 {
-	if (tuner == NULL)
+	return tuner;
+}
+
+Engine* Application::get_engine()
+{
+	return engine;
+}
+
+Streamer& Application::get_streamer()
+{
+	if (streamer == NULL)
 	{
-		throw Exception(_("Tuner has not been initialised"));
+		throw Exception("Streamer has not been initialised");
 	}
 	
-	return *tuner;
+	return *streamer;
 }
 
 void Application::set_dual_language_state(gint state)
 {
-	Engine* engine = main_window->get_engine();
-	if (engine != NULL)
-	{
-		engine->set_dual_language_state(state);
-	}
+	engine->set_dual_language_state(state);
 }
