@@ -18,12 +18,13 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
  */
  
-#include "dvb_si.h"
-#include "log.h"
-#include "string_utility.h"
-#include "application.h"
+#include "dvb_si.hh"
+#include "log.hh"
+#include "string_utility.hh"
+#include "application.hh"
 
 #define SHORT_EVENT			0x4D
+#define EXTENDED_EVENT		0x4E
 
 static guint32 crc_table[256] = {
     0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9, 0x130476dc, 0x17c56b6b,
@@ -84,7 +85,7 @@ static guint32 crc32(const char *data, int len)
     return crc;
 }
 
-ShortEvent::ShortEvent()
+Event::Event()
 {
 	event_id = 0;
 	start_time_MJD = 0;
@@ -113,7 +114,7 @@ DvbSectionParser::DvbSectionParser()
 
 gsize DvbSectionParser::read_section(DvbDemuxer& demuxer)
 {
-	gsize bytes_read = demuxer.read_data(buffer, 3);
+	gsize bytes_read = demuxer.read(buffer, 3);
 	
 	if (bytes_read != 3)
 	{
@@ -122,7 +123,7 @@ gsize DvbSectionParser::read_section(DvbDemuxer& demuxer)
 		
 	gsize remaining_section_length = get_bits (buffer, 12, 12);
 	gsize section_length = remaining_section_length + 3;
-	bytes_read = demuxer.read_data(buffer + 3, remaining_section_length);
+	bytes_read = demuxer.read(buffer + 3, remaining_section_length);
 	
 	if (bytes_read != remaining_section_length)
 	{
@@ -142,7 +143,7 @@ gsize DvbSectionParser::read_section(DvbDemuxer& demuxer)
 void DvbSectionParser::parse_pas(DvbDemuxer& demuxer, ProgramAssociationSection& section)
 {
 	gsize section_length = read_section(demuxer);
-	gsize offset = 8;
+	guint offset = 8;
 	
 	while (offset < (section_length - 4))
 	{
@@ -155,7 +156,7 @@ void DvbSectionParser::parse_pas(DvbDemuxer& demuxer, ProgramAssociationSection&
 	}
 }
 
-int DvbSectionParser::find_descriptor(uint8_t tag, const unsigned char *buf, int descriptors_loop_len, 
+gboolean DvbSectionParser::find_descriptor(uint8_t tag, const unsigned char *buf, int descriptors_loop_len, 
 	const unsigned char **desc, int *desc_len)
 {
 	while (descriptors_loop_len > 0)
@@ -175,70 +176,163 @@ int DvbSectionParser::find_descriptor(uint8_t tag, const unsigned char *buf, int
 
 			if (desc_len)
 				*desc_len = descriptor_len;
-			return 1;
+			return true;
 		}
 
 		buf += descriptor_len;
 		descriptors_loop_len -= descriptor_len;
 	}
-	return 0;
+	
+	return false;
+}
+
+String DvbSectionParser::get_lang_desc(const guchar* buffer)
+{
+	char c[4];
+	String s;
+	memset( mempcpy( c, buffer+2, 3 ), 0, 1 );
+	s = c;
+	return s;
 }
 
 void DvbSectionParser::parse_pms(DvbDemuxer& demuxer, ProgramMapSection& section)
 {	
+	const guchar* desc = NULL;
 	gsize section_length = read_section(demuxer);
-	gsize offset = 8;
-	gsize program_info_len = ((buffer[10] & 0x0f) << 8) | buffer[11];
-	
-	offset += program_info_len + 4;
+	guint offset = 8;
+	gsize program_info_length = ((buffer[10] & 0x0f) << 8) | buffer[11];
+
+	offset += program_info_length + 4;
 	
 	while ((section_length - offset) >= 5)
 	{
 		guint pid_type = buffer[offset];
-		
 		guint elementary_pid = ((buffer[offset+1] & 0x1f) << 8) | buffer[offset+2];
-	    gsize descriptor_len = ((buffer[offset+3] & 0x0f) << 8) | buffer[offset+4];
+	    gsize descriptor_length = ((buffer[offset+3] & 0x0f) << 8) | buffer[offset+4];
 		
 		switch (pid_type)
 		{
-		case 0x01:
-		case 0x02:
+		case STREAM_TYPE_MPEG1:
+		case STREAM_TYPE_MPEG2:
+		case STREAM_TYPE_MPEG4:
+		case STREAM_TYPE_H264:
 			Log::write(_("Video PID: %d"), elementary_pid);
-			section.video_pids.push_back(elementary_pid);
+			{
+				VideoStream video_stream;
+				video_stream.pid = elementary_pid;
+				video_stream.type = pid_type;
+				section.video_streams.push_back(video_stream);
+			}
 			break;
 		case 0x03:
 		case 0x04:
 			Log::write(_("Audio PID: %d"), elementary_pid);
-			section.audio_pids.push_back(elementary_pid);
+			{				
+				AudioStream stream;
+				stream.pid = elementary_pid;
+				stream.is_ac3 = false;
+
+				if (find_descriptor(0x0A, buffer + offset + 5, descriptor_length, &desc, NULL))
+				{
+					stream.language = get_lang_desc (desc);
+				}
+
+				section.audio_streams.push_back(stream);
+			}
 			break;
 		case 0x06:
-			if (find_descriptor(0x56, buffer + offset + 5, descriptor_len, NULL, NULL))
 			{
-				Log::write(_("Teletext PID: %d"), elementary_pid);
-				section.teletext_pids.push_back(elementary_pid);
-			}
-			else if (find_descriptor (0x59, buffer + offset + 5, descriptor_len, NULL, NULL))
-			{
-				Log::write(_("Subtitle PID: %d"), elementary_pid);
-				section.subtitle_pids.push_back(elementary_pid);				
-			}
-			else if (find_descriptor (0x6a, buffer + offset + 5, descriptor_len, NULL, NULL))
-			{
-				Log::write(_("AC3 PID: %d"), elementary_pid);
-				section.ac3_pids.push_back(elementary_pid);				
+				int descriptor_loop_length = 0;
+				if (find_descriptor(0x56, buffer + offset + 5, descriptor_length, &desc, &descriptor_loop_length))
+				{
+					Log::write(_("Teletext PID: %d"), elementary_pid);
+
+					TeletextStream stream;
+					stream.pid = elementary_pid;
+
+					int descriptor_index = 0;
+					while (descriptor_index < descriptor_loop_length - 4)
+					{
+						TeletextLanguageDescriptor descriptor;
+						
+						descriptor.language			= get_lang_desc(desc + descriptor_index);
+						descriptor.type				= get_bits(desc + descriptor_index + 6, 0, 5);
+						descriptor.magazine_number	= get_bits(desc + descriptor_index + 6, 5, 3);
+						descriptor.page_number		= get_bits(desc + descriptor_index + 7, 0, 8);
+						
+						Log::write("TeleText: Language: '%s', Type: %d, Magazine Number: %d, Page Number: %d",
+							descriptor.language.c_str(),
+							descriptor.type,
+							descriptor.magazine_number,
+							descriptor.page_number);
+						
+						stream.languages.push_back(descriptor);
+						
+						descriptor_index += 5;
+					}
+					section.teletext_streams.push_back(stream);
+				}
+				else if (find_descriptor (0x59, buffer + offset + 5, descriptor_length, &desc, NULL))
+				{
+					Log::write(_("Subtitle PID: %d"), elementary_pid);
+					SubtitleStream stream;
+					stream.pid = elementary_pid;
+					if (descriptor_length > 0)
+					{
+						stream.subtitling_type = get_bits(desc + 5, 0, 8);
+						if (stream.subtitling_type>=0x10 && stream.subtitling_type<=0x23)
+						{
+							stream.language				= get_lang_desc(desc);
+							stream.composition_page_id	= get_bits(desc + 6, 0, 16);
+							stream.ancillary_page_id	= get_bits(desc + 8, 0, 16);
+							
+							Log::write(
+								"Subtitle composition_page_id: %d, ancillary_page_id: %d, language: %s",
+								stream.composition_page_id, stream.ancillary_page_id, stream.language.c_str() );
+						}
+					}
+					section.subtitle_streams.push_back(stream);
+				}
+				else if (find_descriptor (0x6A, buffer + offset + 5, descriptor_length, NULL, NULL))
+				{
+					Log::write(_("AC3 PID: %d"), elementary_pid);
+					AudioStream stream;
+					stream.pid = elementary_pid;
+					stream.is_ac3 = true;
+
+					const guchar* desc = NULL;
+					if (find_descriptor(0x0A, buffer + offset + 5, descriptor_length, &desc, NULL))
+					{
+						stream.language = get_lang_desc (desc);
+					}
+					section.audio_streams.push_back(stream);
+				}
 			}
 			break;
 		case 0x81:
 			Log::write(_("AC3 PID: %d"), elementary_pid);
-			section.ac3_pids.push_back(elementary_pid);
+			{
+				AudioStream stream;
+				stream.pid = elementary_pid;
+				stream.is_ac3 = true;
+			
+				const guchar* desc = NULL;
+				if (find_descriptor(0x0A, buffer + offset + 5, descriptor_length, &desc, NULL))
+				{
+					stream.language = get_lang_desc (desc);
+				}
+				section.audio_streams.push_back(stream);
+			}
+			break;
+		default:
 			break;
 		}
 		
-		offset += descriptor_len + 5;
+		offset += descriptor_length + 5;
 	}
 }
 
-void DvbSectionParser::parse_eis(DvbDemuxer& demuxer, DvbEventInformationSection& section)
+void DvbSectionParser::parse_eis(DvbDemuxer& demuxer, EventInformationSection& section)
 {
 	gsize section_length = read_section(demuxer);
 	
@@ -258,42 +352,42 @@ void DvbSectionParser::parse_eis(DvbDemuxer& demuxer, DvbEventInformationSection
 	
 	while (offset < (section_length - CRC_BYTE_SIZE))
 	{
-		ShortEvent short_event;
+		Event event;
 
-		short_event.event_id                   = get_bits (&buffer[offset], 0, 16);
-		short_event.start_time_MJD             = get_bits (&buffer[offset], 16, 16);
-		short_event.start_time_UTC             = get_bits (&buffer[offset], 32, 24);
-		short_event.duration                   = get_bits (&buffer[offset], 56, 24);
-		short_event.running_status             = get_bits (&buffer[offset], 80, 3);
-		short_event.free_CA_mode               = get_bits (&buffer[offset], 83, 1);
+		event.event_id                   = get_bits (&buffer[offset], 0, 16);
+		event.start_time_MJD             = get_bits (&buffer[offset], 16, 16);
+		event.start_time_UTC             = get_bits (&buffer[offset], 32, 24);
+		event.duration                   = get_bits (&buffer[offset], 56, 24);
+		event.running_status             = get_bits (&buffer[offset], 80, 3);
+		event.free_CA_mode               = get_bits (&buffer[offset], 83, 1);
 
 		unsigned int descriptors_loop_length  = get_bits (&buffer[offset], 84, 12);
 		offset += 12;
 		unsigned int end_descriptor_offset = descriptors_loop_length + offset;
 
-		short_event.dur_hour = ((short_event.duration >> 20)&0xf)*10+((short_event.duration >> 16)&0xf);
-		short_event.dur_min =  ((short_event.duration >> 12)&0xf)*10+((short_event.duration >>  8)&0xf);
-		short_event.dur_sec =  ((short_event.duration >>  4)&0xf)*10+((short_event.duration      )&0xf);
+		event.dur_hour = ((event.duration >> 20)&0xf)*10+((event.duration >> 16)&0xf);
+		event.dur_min =  ((event.duration >> 12)&0xf)*10+((event.duration >>  8)&0xf);
+		event.dur_sec =  ((event.duration >>  4)&0xf)*10+((event.duration      )&0xf);
 
-		if ( short_event.start_time_MJD > 0 )
+		if ( event.start_time_MJD > 0 )
 		{
-			long year =  (long) ((short_event.start_time_MJD  - 15078.2) / 365.25);
-			long month =  (long) ((short_event.start_time_MJD - 14956.1 - (long)(year * 365.25) ) / 30.6001);
-			short_event.start_day =  (long) (short_event.start_time_MJD - 14956 - (long)(year * 365.25) - (long)(month * 30.6001));
+			long year =  (long) ((event.start_time_MJD  - 15078.2) / 365.25);
+			long month =  (long) ((event.start_time_MJD - 14956.1 - (long)(year * 365.25) ) / 30.6001);
+			event.start_day =  (long) (event.start_time_MJD - 14956 - (long)(year * 365.25) - (long)(month * 30.6001));
 
 			long k = (month == 14 || month == 15) ? 1 : 0;
 
-			short_event.start_year = year + k + 1900;
-			short_event.start_month = month - 1 - k * 12;
+			event.start_year = year + k + 1900;
+			event.start_month = month - 1 - k * 12;
 		}
 
-		short_event.start_hour = ((short_event.start_time_UTC >> 20)&0xf)*10+((short_event.start_time_UTC >> 16)&0xf);
-		short_event.start_min =  ((short_event.start_time_UTC >> 12)&0xf)*10+((short_event.start_time_UTC >>  8)&0xf);
-		short_event.start_sec =  ((short_event.start_time_UTC >>  4)&0xf)*10+((short_event.start_time_UTC      )&0xf);
+		event.start_hour = ((event.start_time_UTC >> 20)&0xf)*10+((event.start_time_UTC >> 16)&0xf);
+		event.start_min =  ((event.start_time_UTC >> 12)&0xf)*10+((event.start_time_UTC >>  8)&0xf);
+		event.start_sec =  ((event.start_time_UTC >>  4)&0xf)*10+((event.start_time_UTC      )&0xf);
 
 		while (offset < end_descriptor_offset)
 		{
-			offset += decode_short_event(&buffer[offset], short_event);
+			offset += decode_short_event(&buffer[offset], event);
 		}
 		
 		if (offset > end_descriptor_offset)
@@ -301,7 +395,7 @@ void DvbSectionParser::parse_eis(DvbDemuxer& demuxer, DvbEventInformationSection
 			throw Exception("ASSERT: offset > end_descriptor_offset");
 		}
 
-		section.short_events.push_back( short_event );
+		section.events.push_back( event );
 	}
 	section.crc = get_bits (&buffer[offset], 0, 32);
 	offset += 4;
@@ -312,7 +406,7 @@ void DvbSectionParser::parse_eis(DvbDemuxer& demuxer, DvbEventInformationSection
 	}
 }
 
-gsize DvbSectionParser::decode_short_event (const guchar* buffer, ShortEvent& descriptor)
+gsize DvbSectionParser::decode_short_event (const guchar* buffer, Event& event)
 {
 	if (buffer[1] == 0)
 	{
@@ -325,29 +419,31 @@ gsize DvbSectionParser::decode_short_event (const guchar* buffer, ShortEvent& de
 	{
 	case SHORT_EVENT:
 		{
+			String language;
 			String title;
 			String description;
 			
+			language = get_lang_desc (buffer);
 			unsigned int offset = 5;
 			offset += get_text(title, &buffer[offset]);
 			offset += get_text(description, &buffer[offset]);
 			
-			if (descriptor.title.is_empty())
+			if (event.title.is_empty())
 			{
-				descriptor.title = title;
+				event.title = title;
 			}
-			else if (descriptor.title != title)
+			else if (event.title != title)
 			{
-				descriptor.title += "-" + title;
+				event.title += "-" + title;
 			}
 			
-			if (descriptor.description.is_empty())
+			if (event.description.is_empty())
 			{
-				descriptor.description = description;
+				event.description = description;
 			}
-			else if (descriptor.description != description)
+			else if (event.description != description)
 			{
-				descriptor.description += "-" + description;
+				event.description += "-" + description;
 			}
 		}
 		break;
@@ -405,6 +501,7 @@ gsize DvbSectionParser::get_text(String& s, const guchar* buffer)
 				case 0x09: codeset = "ISO-8859-13"; break;
 				case 0x0A: codeset = "ISO-8859-14"; break;
 				case 0x0B: codeset = "ISO-8859-15"; break;
+				case 0x14: codeset = "UTF-16BE"; break;
 
 				case 0x10:
 					{
@@ -456,21 +553,28 @@ gsize DvbSectionParser::get_text(String& s, const guchar* buffer)
 				{
 					u_char ch = buffer[index];
 
-					if (ch == 0x86 || ch == 0x87)
+					if (codeset == "UTF-16BE")
 					{
-						// Ignore formatting
-					}
-					else if (ch == 0x8A)
-					{
-						text[text_index++] = '\n';
-					}
-					else if (ch >= 0x80 && ch < 0xA0)
-					{
-						text[text_index++] = '.';
+						text[text_index++] = ch;
 					}
 					else
 					{
-						text[text_index++] = ch;
+						if (ch == 0x86 || ch == 0x87)
+						{
+							// Ignore formatting
+						}
+						else if (ch == 0x8A)
+						{
+							text[text_index++] = '\n';
+						}
+						else if (ch >= 0x80 && ch < 0xA0)
+						{
+							text[text_index++] = '.';
+						}
+						else
+						{
+							text[text_index++] = ch;
+						}
 					}
 					
 					index++;
@@ -488,7 +592,7 @@ gsize DvbSectionParser::get_text(String& s, const guchar* buffer)
 					&bytes_read,
 					&bytes_written,
 					&error);
-				
+
 				if (error != NULL)
 				{
 					String message = String::format(_("Failed to convert to UTF-8: %s"), error->message);
