@@ -18,7 +18,7 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
  */
  
-#include "engine.h"
+#include "engine.hh"
 
 Engine::Engine()
 {
