@@ -18,10 +18,10 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
  */
  
-#include "application.h"
-#include "exception_handler.h"
-#include "gdk_lock.h"
-#include "xine_engine.h"
+#include "application.hh"
+#include "exception_handler.hh"
+#include "gdk_lock.hh"
+#include "xine_engine.hh"
 #include "config.h"
 
 int main (int argc, char** argv)
@@ -29,7 +29,7 @@ int main (int argc, char** argv)
 	TRY
 
 	XineEngine::initialise();
-	
+
 #ifdef ENABLE_NLS
 	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
