@@ -59,13 +59,13 @@
 #include <linux/dvb/frontend.h>
 #include <linux/dvb/dmx.h>
 
-#include "scan.h"
-#include "exception.h"
-#include "dvb_si.h"
-#include "string_utility.h"
-#include "io.h"
-#include "log.h"
-#include "application.h"
+#include "scan.hh"
+#include "exception.hh"
+#include "dvb_si.hh"
+#include "string_utility.hh"
+#include "io.hh"
+#include "log.hh"
+#include "application.hh"
 #include <libgnome/libgnome.h>
 
 ScanDialog* scan_dialog = NULL;
@@ -212,8 +212,8 @@ static __inline__ int list_empty(struct list_head *head)
 #define PACKED __attribute((packed))
 
 struct lnb_types_st {
-	const char	*name;
-	const char	**desc;
+	const char		*name;
+	const char		**desc;
 	unsigned long	low_val;
 	unsigned long	high_val;	/* zero indicates no hiband */
 	unsigned long	switch_val;	/* zero indicates no hiband */
@@ -369,12 +369,13 @@ static const char *cmulti_desc[] = {
 		(char *)NULL };
 
 static struct lnb_types_st lnbs[] = {
-	{"UNIVERSAL",	univ_desc,		9750, 10600, 11700 },
- 	{"DBS",		dbs_desc, 		11250, 0, 0 },
+	{"UNIVERSAL",	univ_desc,			9750, 10600, 11700 },
+ 	{"DBS",			dbs_desc, 			11250, 0, 0 },
 	{"STANDARD",	standard_desc,		10000, 0, 0 },
-	{"ENHANCED",	enhan_desc,		9750, 0, 0 },
-	{"C-BAND",	cband_desc,		5150, 0, 0 },
-	{"C-MULTI",	cmulti_desc,		5150, 5750, 0 }
+	{"ENHANCED",	enhan_desc,			9750, 0, 0 },
+	{"C-BAND",		cband_desc,			5150, 0, 0 },
+	{"C-MULTI",		cmulti_desc,		5150, 5750, 0 },
+	{"NZ-STANDARD", NULL,				11300, 0, 0 }
 };
 
 /* shamelessly stolen from dvbsnoop, but modified */
@@ -517,18 +518,21 @@ lnb_enum(int curno)
  * otherwise low[,high[,switch]]
  */
 
-int
-lnb_decode(char *str, struct lnb_types_st *lnbp)
+int lnb_decode(const char *str, struct lnb_types_st *lnbp)
 {
-int i;
-char *cp, *np;
+	int i;
+	const char *cp = NULL;
+	char *np;
 
 	memset(lnbp, 0, sizeof(*lnbp));
 	cp = str;
 	while(*cp && isspace(*cp))
 		cp++;
-	if (isalpha(*cp)) {
-		for (i = 0; i < (int) (sizeof(lnbs) / sizeof(lnbs[0])); i++) {
+	
+	if (isalpha(*cp))
+	{
+		for (i = 0; i < (int) (sizeof(lnbs) / sizeof(lnbs[0])); i++)
+		{
 			if (!strcasecmp(lnbs[i].name, cp)) {
 				*lnbp = lnbs[i];
 				return 1;
@@ -536,6 +540,7 @@ char *cp, *np;
 		}
 		return -1;
 	}
+	
 	if (*cp == '\0' || !isdigit(*cp))
 		return -1;
 	lnbp->low_val = strtoul(cp, &np, 0);
@@ -557,6 +562,7 @@ char *cp, *np;
 	if (!isdigit(*cp))
 		return -1;
 	lnbp->switch_val = strtoul(cp, NULL, 0);
+
 	return 1;
 }
 
@@ -690,7 +696,7 @@ static int no_ATSC_PSIP;
 static int ATSC_type=1;
 static int ca_select = 1;
 static int serv_select = 7;
-static struct lnb_types_st lnb_type;
+static struct lnb_types_st lnb;
 static int unique_anon_services;
 
 static enum fe_spectral_inversion spectral_inversion = INVERSION_AUTO;
@@ -2086,8 +2092,8 @@ static int __tune_to_transponder (int frontend_fd, struct transponder *t)
 	if (t->type == FE_QPSK) {
 		int hiband = 0;
 
-		if (lnb_type.switch_val && lnb_type.high_val &&
-			p.frequency >= lnb_type.switch_val)
+		if (lnb.switch_val && lnb.high_val &&
+			p.frequency >= lnb.switch_val)
 			hiband = 1;
 
 		setup_switch (frontend_fd,
@@ -2096,9 +2102,9 @@ static int __tune_to_transponder (int frontend_fd, struct transponder *t)
 			      hiband);
 		usleep(50000);
 		if (hiband)
-			p.frequency = abs(p.frequency - lnb_type.high_val);
+			p.frequency = abs(p.frequency - lnb.high_val);
 		else
-			p.frequency = abs(p.frequency - lnb_type.low_val);
+			p.frequency = abs(p.frequency - lnb.low_val);
 	}
 
 	if (ioctl(frontend_fd, FE_SET_FRONTEND, &p) == -1) {
@@ -2580,16 +2586,25 @@ static void dump_lists (IO::Channel& out_file)
 
 int scan_main (const gchar* region_file_path, ScanDialog* sd)
 {
+	Application& application = Application::get_current();
+	Configuration& configuration = application.get_configuration();
+
 	int frontend_fd;
 	scan_dialog = sd;
-	serv_select = 1;
 	
 	/* start with default lnb type */
-	lnb_type = *lnb_enum(0);
+	lnb = *lnb_enum(0);
+
+	String lnb_type = configuration.get_string_value("lnb_type");
+	if (lnb_type != "none")
+	{
+		lnb.low_val *= 1000;	/* convert to kiloherz */
+		lnb.high_val *= 1000;	/* convert to kiloherz */
+		lnb.switch_val *= 1000;	/* convert to kiloherz */
+				
+		lnb_decode(lnb_type.c_str(), &lnb);	
+	}
 
-	lnb_type.low_val *= 1000;	/* convert to kiloherz */
-	lnb_type.high_val *= 1000;	/* convert to kiloherz */
-	lnb_type.switch_val *= 1000;	/* convert to kiloherz */
 	if (switch_pos >= 4)
 	{
 		throw Exception("Switch position needs to be < 4!");
@@ -2600,12 +2615,10 @@ int scan_main (const gchar* region_file_path, ScanDialog* sd)
 		poll_fds[i].fd = -1;
 	}
 
-	Application& application = Application::get_current();
-	DvbTuner& tuner = application.get_tuner();
-	frontend_fd = tuner.get_fd();
-	tuner.get_frontend_info(&fe_info);
+	DvbTuner* tuner = application.get_tuner();
+	frontend_fd = tuner->get_fd();
+	tuner->get_frontend_info(&fe_info);
 	
-	Configuration& configuration = application.get_configuration();
 	demux_path = configuration.get_string_value("demux_path");
 	String channel_file_path = configuration.get_string_value("channel_file_path");
 	
