@@ -18,8 +18,8 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
  */
  
-#include "transponder.h"
-#include "log.h"
+#include "transponder.hh"
+#include "log.hh"
 
 Transponder::Transponder()
 {
